/*
 *  kexecboot - A kexec based bootloader
 *
 *  Copyright (c) 2008-2009 Yuri Bushmelev <jay4mail@gmail.com>
 *  Copyright (c) 2008 Thomas Kunze <thommycheck@gmx.de>
 *  Copyright (c) 2006 Matthew Allum <mallum@o-hand.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */
#include "fb.h"

void fb_render(FB * fb)
{
	/* klibc uses 8bit transfers that breaks image on tosa */
	/* memcpy(fb->data, fb->backbuffer, fb->screensize); */
	uint16 *source, *dest;
	int n = fb->screensize/2;

	source = (uint16 *)fb->backbuffer;
	dest = (uint16 *)fb->data;

	while (n--) {
		*dest++ = *source++;
	}
}

void fb_destroy(FB * fb)
{
	if (fb->fd >= 0)
		close(fb->fd);
	if(fb->backbuffer)
		free(fb->backbuffer);
	free(fb);
}

static int
attempt_to_change_pixel_format(FB * fb, struct fb_var_screeninfo *fb_var)
{
	/* By default the framebuffer driver may have set an oversized
	 * yres_virtual to support VT scrolling via the panning interface.
	 *
	 * We don't try and maintain this since it's more likely that we
	 * will fail to increase the bpp if the driver's pre allocated
	 * framebuffer isn't large enough.
	 */
	fb_var->yres_virtual = fb_var->yres;

	/* First try setting an 8,8,8,0 pixel format so we don't have to do
	 * any conversions while drawing. */

	fb_var->bits_per_pixel = 32;

	fb_var->red.offset = 0;
	fb_var->red.length = 8;

	fb_var->green.offset = 8;
	fb_var->green.length = 8;

	fb_var->blue.offset = 16;
	fb_var->blue.length = 8;

	fb_var->transp.offset = 0;
	fb_var->transp.length = 0;

	if (ioctl(fb->fd, FBIOPUT_VSCREENINFO, fb_var) == 0) {
		fprintf(stdout,
			"Switched to a 32 bpp 8,8,8 frame buffer\n");
		return 1;
	} else {
		fprintf(stderr,
			"Error, failed to switch to a 32 bpp 8,8,8 frame buffer\n");
	}

	/* Otherwise try a 16bpp 5,6,5 format */

	fb_var->bits_per_pixel = 16;

	fb_var->red.offset = 11;
	fb_var->red.length = 5;

	fb_var->green.offset = 5;
	fb_var->green.length = 6;

	fb_var->blue.offset = 0;
	fb_var->blue.length = 5;

	fb_var->transp.offset = 0;
	fb_var->transp.length = 0;

	if (ioctl(fb->fd, FBIOPUT_VSCREENINFO, fb_var) == 0) {
		fprintf(stdout,
			"Switched to a 16 bpp 5,6,5 frame buffer\n");
		return 1;
	} else {
		fprintf(stderr,
			"Error, failed to switch to a 16 bpp 5,6,5 frame buffer\n");
	}

	return 0;
}

FB *fb_new(int angle)
{
	struct fb_var_screeninfo fb_var;
	struct fb_fix_screeninfo fb_fix;
	int off;
	char *fbdev;

	FB *fb = NULL;

	fbdev = getenv("FBDEV");
	if (fbdev == NULL)
		fbdev = "/dev/fb0";

	if ((fb = malloc(sizeof(FB))) == NULL) {
		perror("Error no memory");
		goto fail;
	}

	memset(fb, 0, sizeof(FB));

	fb->fd = -1;

	if ((fb->fd = open(fbdev, O_RDWR)) < 0) {
		perror("Error opening /dev/fb0");
		goto fail;
	}

	if (ioctl(fb->fd, FBIOGET_VSCREENINFO, &fb_var) == -1) {
		perror("Error getting variable framebuffer info");
		goto fail;
	}

	if (fb_var.bits_per_pixel != 1 && fb_var.bits_per_pixel != 2
		&& fb_var.bits_per_pixel < 16)
	{
		fprintf(stderr,
			"Error, no support currently for %i bpp frame buffers\n"
			"Trying to change pixel format...\n",
			fb_var.bits_per_pixel);
		if (!attempt_to_change_pixel_format(fb, &fb_var))
			goto fail;
	}

	/* NB: It looks like the fbdev concept of fixed vs variable screen info is
	 * broken. The line_length is part of the fixed info but it can be changed
	 * if you set a new pixel format. */
	if (ioctl(fb->fd, FBIOGET_FSCREENINFO, &fb_fix) == -1) {
		perror("Error getting fixed framebuffer info");
		goto fail;
	}

	fb->real_width = fb->width = fb_var.xres;
	fb->real_height = fb->height = fb_var.yres;
	fb->bpp = fb_var.bits_per_pixel;
	fb->stride = fb_fix.line_length;
	fb->type = fb_fix.type;
	fb->visual = fb_fix.visual;
	fb->screensize = fb->width * fb->height * fb->bpp/8;
	fb->backbuffer = malloc(fb->screensize);

	fb->base = (char *) mmap((caddr_t) NULL,
				 /*fb_fix.smem_len */
				 fb->stride * fb->height,
				 PROT_READ | PROT_WRITE,
				 MAP_SHARED, fb->fd, 0);

	if (fb->base == (char *) -1) {
		perror("Error cannot mmap framebuffer ");
		goto fail;
	}

	off =
	    (unsigned long) fb_fix.smem_start %
	    (unsigned long) getpagesize();

	fb->data = fb->base + off;
#if 0
	/* FIXME: No support for 8pp as yet  */
	if (visual == FB_VISUAL_PSEUDOCOLOR
	    || visual == FB_VISUAL_STATIC_PSEUDOCOLOR) {
		static struct fb_cmap cmap;

		cmap.start = 0;
		cmap.len = 16;
		cmap.red = saved_red;
		cmap.green = saved_green;
		cmap.blue = saved_blue;
		cmap.transp = NULL;

		ioctl(fb, FBIOGETCMAP, &cmap);
	}

	if (!status)
		atexit(bogl_done);
	status = 2;
#endif

	fb->angle = angle;

	switch (fb->angle) {
	case 270:
	case 90:
		fb->width = fb->real_height;
		fb->height = fb->real_width;
		break;
	case 180:
	case 0:
	default:
		break;
	}

	return fb;

      fail:

	if (fb)
		fb_destroy(fb);

	return NULL;
}

static inline int
fb_offset(FB *fb, int x, int y)
{
	switch (fb->bpp)
	{
	/* pixel offset */
	case 2: return (y * (fb->stride << 2)) + x;
	case 1: return (y * (fb->stride << 3)) + x;
	/* byte offset */
	default: return (y * fb->stride) + (x * (fb->bpp >> 3));
	}
}

inline void
fb_plot_pixel(FB * fb, int x, int y, uint8 red, uint8 green, uint8 blue)
{
	int off, shift;

	if (x < 0 || x > (fb->width - 1) || y < 0 || y > (fb->height - 1))
		return;

	switch (fb->angle) {
	case 270:
		off = fb_offset(fb, fb->height - y - 1, x);
		break;
	case 180:
		off = fb_offset(fb, fb->width - x - 1, fb->height - y - 1);
		break;
	case 90:
		off = fb_offset(fb, y, fb->width - x - 1);
		break;
	case 0:
	default:
		off = fb_offset(fb, x, y);
		break;
	}

	/* FIXME: handle no RGB orderings */
	switch (fb->bpp) {
	case 24:
	case 32:
		*(fb->backbuffer + off) = red;
		*(fb->backbuffer + off + 1) = green;
		*(fb->backbuffer + off + 2) = blue;
	//	*(fb->data + off) = red;
	//	*(fb->data + off + 1) = green;
	//	*(fb->data + off + 2) = blue;
		break;
	case 16:
	//	*(volatile uint16 *) (fb->data + off)
		*(volatile uint16 *) (fb->backbuffer + off)
		    = ((red >> 3) << 11) | ((green >> 2) << 5) | (blue >> 3);
		break;
	case 2:
		shift = (3 - (off & 3)) << 1;
		*(fb->data + (off >> 2)) = (*(fb->data + (off >> 2)) & ~(3 << shift))
		| (((11*red + 16*green + 5*blue) >> 11) << shift);
		break;
	case 1:
		shift = 7 - (off & 7);
		if (((11*red + 16*green + 5*blue) >> 5) >= 128)
			*(fb->data + (off >> 3)) |= (1 << shift);
		else
			*(fb->data + (off >> 3)) &= ~(1 << shift);
		break;
	default:
		/* depth not supported yet */
		break;
	}

}

void fb_draw_rect(FB * fb, int x, int y, int width, int height,
		uint8 red, uint8 green, uint8 blue)
{
	int dx, dy;

	for (dy = 0; dy < height; dy++)
		for (dx = 0; dx < width; dx++)
			fb_plot_pixel(fb, x + dx, y + dy, red, green,
				      blue);
}

void fb_draw_image(FB * fb, int x, int y, int img_width, int img_height,
		int img_bytes_per_pixel, uint8 * rle_data)
{
	uint8 *p = rle_data;
	int dx = 0, dy = 0, total_len;
	unsigned int len;

	total_len = img_width * img_height * img_bytes_per_pixel;

	/* FIXME: Optimise, check for over runs ... */
	while ((p - rle_data) < total_len) {
		len = *(p++);

		if (len & 128) {
			len -= 128;

			if (len == 0)
				break;

			do {
				fb_plot_pixel(fb, x + dx, y + dy, *p,
					      *(p + 1), *(p + 2));
				if (++dx >= img_width) {
					dx = 0;
					dy++;
				}
			}
			while (--len && (p - rle_data) < total_len);

			p += img_bytes_per_pixel;
		} else {
			if (len == 0)
				break;

			do {
				fb_plot_pixel(fb, x + dx, y + dy, *p,
					      *(p + 1), *(p + 2));
				if (++dx >= img_width) {
					dx = 0;
					dy++;
				}
				p += img_bytes_per_pixel;
			}
			while (--len && (p - rle_data) < total_len);
		}
	}
}

/* Font rendering code based on BOGL by Ben Pfaff */

static int font_glyph(const Font * font, char wc, u_int32_t ** bitmap)
{
	int mask = font->index_mask;
	int i;

	for (;;) {
		for (i = font->offset[wc & mask]; font->index[i]; i += 2) {
			if ((font->index[i] & ~mask) == (wc & ~mask)) {
				if (bitmap != NULL)
					*bitmap =
					    &font->content[font->
							   index[i + 1]];
				return font->index[i] & mask;
			}
		}
	}
	return 0;
}

void fb_text_size(FB * fb, int *width, int *height, const Font * font,
		const char *text)
{
	char *c = (char *) text;
	int n, w, h, mw;

	n = strlen(text);
	mw = h = w = 0;

	for(;*c;c++){
		if (*c == '\n') {
			if (w > mw)
				mw = 0;
			h += font->height;
			continue;
		}

		w += font_glyph(font, *c, NULL);
	}

	*width = (w > mw) ? w : mw;
	*height = (h == 0) ? font->height : h;
}

void fb_draw_text(FB * fb, int x, int y, uint8 red, uint8 green, uint8 blue,
		const Font * font, const char *text)
{
	int h, w, n, cx, cy, dx, dy;
	char *c = (char *) text;

	n = strlen(text);
	h = font->height;
	dx = dy = 0;

	for(; *c;c++){
		u_int32_t *glyph = NULL;

		if (*c == '\n') {
			dy += h;
			dx = 0;
			continue;
		}

		w = font_glyph(font, *c, &glyph);

		if (glyph == NULL)
			continue;

		for (cy = 0; cy < h; cy++) {
			u_int32_t g = *glyph++;

			for (cx = 0; cx < w; cx++) {
				if (g & 0x80000000)
					fb_plot_pixel(fb, x + dx + cx,
						      y + dy + cy, red,
						      green, blue);
				g <<= 1;
			}
		}

		dx += w;
	}
}
